#include <iostream>
#include "phAdapt.h"
#include "shapeFunction.h"

#include <stdio.h>
#include <stdlib.h>

using namespace std;

extern pMeshDataId phasta_solution;

// called in callback = phastaTransfer
double* 
InterpolateSolution( pRegion region, 
                     double xi[3], 
                     int ndof,
                     pMeshDataId modes ) {

/* This routine takes the region and the location of the vertex in */
/* parametric coordinates, polynomial order, and the attached solution*/
/* and returns the solution interpolated at the specified location. Higher*/
/* order mode coefficients are also evaluated    */
  int exitCode=0;
  double* vcc;
  double* ecc;
  double* fcc;
  double sfval;
  double* q = new double[ndof];
  double r=xi[0];
  double s=xi[1];
  double t=xi[2];
  double L[4] = { r, s, t, 1-r-s-t};
  
  for (int i=0; i < ndof; i++) q[i] = 0.0;
  
  pVertex vrts[4];
  // pEdge edgs[6];
  // pFace facs[4];
  
  pPList rverts = R_vertices(region,1);
  for(int iVert=0; iVert<PList_size(rverts); iVert++) {
      vrts[iVert] = (pVertex)PList_item(rverts,iVert);
  }
  PList_delete(rverts);
  
  // loop over the vertices of (parent) region
  for (int k=0; k<4; k++) {
    
    // solution(( pEntity )vrts[k], phasta_solution, &vcc ); 
    /* get the solution at the vertex */

    if(!EN_getDataPtr( (pEntity)vrts[k], phasta_solution, (void**)&vcc )) {
      V_info(vrts[k]);
      printf("\nerror in InterpolateSolution: wanted to retrieve "
	     "a solution on a vertex that is empty\n");
      exit(-1);
    }

    /*evaluate the shape function at the given location xi*/
    sfval = Entity_shapeFunction(( pEntity )vrts[k],
				 ( pEntity )region, 1, 0, L);
    for (int i=0; i < ndof; i++){
      /*sigma(N_a(xi)*y_a)*/
      q[i] += vcc[i]*sfval;
    }
  }
  
  // get the edge and face coefficients
  // not implemented at the moment
  //      for (int k=0; k<6; k++) {// loop over edges
  //          int nem =0;
  //          solution(( pEntity )edgs[k], phSol, &ecc ); 
  //          numberofmodes((pEntity)edgs[k], modes, &nem );
  //          if ( ecc ) {
  //              for( int e=0; e < nem ; e++ ) {
  //                  sfval = Entity_shapeFunction((pEntity)edgs[k],
  //                                               (pEntity)region,e+2,0,L);
  //                  for (int j=0; j < ndof; j++) 
  //                      q[j] += ecc[e*ndof+j]*sfval;
  //              }
  //          }
  //      }
  //      for (int k=0; k<4; k++) {// loop over faces
  //          int nem=0;
  //          solution(( pEntity )facs[k], phSol, &ecc ); 
  //          numberofmodes((pEntity)facs[k], modes, &nem );
  //          if ( ecc ) {
  //              for( int e=0; e < nem ; e++ ) {
  //                  sfval = Entity_shapeFunction((pEntity)facs[k],
  //                                               (pEntity)region,e+3,0,L);
  //                  for (int j=0; j < ndof; j++) 
  //                      q[j] += ecc[e*ndof+j]*sfval;
  //              }
  //          }
  //      }
  return q;
}
