#include <iostream>
#include "phAdapt.h"

using namespace std;


extern pMeshDataId nodalgradientID;
extern pMeshDataId nodalhessianID;

void 
ModifyHessiansAtBdry(pMesh mesh) {
  pVertex v;
  VIter vIter=M_vertexIter(mesh);
  while(v = VIter_next(vIter)) {
    
    // bdry correction:
    // take values of (any/closest) OFF-bdry vertex as the bdry vertex value
    // all verts classified on model :face, edge or vertex
    if(V_whatInType(v) != 3) {      
      int numEdges = V_numEdges(v);
      pEdge edge;
      bool finished=false;
      pVertex otherVertex;
      for(int i=0; i<numEdges  ;i++){
	edge = V_edge(v,i);
	
	// if the edge is in interior
	// (later distance could be added) 
	if(E_whatInType(edge) == 3){
	  otherVertex = E_otherVertex(edge,v);
	  finished=true;
	  break;
	}
      }

      if(finished) {
	
	double *nodalHessianOther;
	if(!EN_getDataPtr((pEntity)otherVertex, nodalhessianID,(void**)&nodalHessianOther)){
	  
	  cout<<"\nerror in :V_Hessian no data attached to OTHER vertex\n";
	  V_info(otherVertex);
	  exit(0);
	}
	
	double *nodalHessianActual;
	if(!EN_getDataPtr((pEntity)v, nodalhessianID,(void**)&nodalHessianActual)){
	  
	  cout<<"\nerror in :V_Hessian no data attached to ACTUAL vertex\n";
	  exit(0);
	}
	
	delete [] nodalHessianActual;
	double* nodalHessianModified=new double[6];
	
	for(int iHess=0;iHess<6;iHess++) {
	  nodalHessianModified[iHess]=nodalHessianOther[iHess];
	}
	
	EN_deleteData((pEntity)v, nodalhessianID);
	
	EN_attachDataPtr( (pEntity)v, nodalhessianID, (void *)
			  nodalHessianModified );
      }
    }
    // one comment is appropriate:
    // if the bdry vertex does NOT have any edges
    // which are interior (very pathologigal mesh anyway)
    // the hessian on that node is kept as it is 
    // (mesh should not have such pathological case)
  }
  VIter_delete(vIter); 
}
  
