#include <iostream>
#include "phAdapt.h"

using namespace std;



extern pMeshDataId phasta_solution;

// Modify mesh-metric to take special care
// like in parallel plates (i.e, w=f(y)) 
// if a user wants different hmax in x-direction
// other examples can be situations where
// user don't want to have one mesh edge 
// connecting two geometric model edges (like no dofs)
void ModifyMetric(pVertex vertex, double dir[3][3], double* h){
  double tol = 1.e-2; 
  double* nodalSolution;

  pVertex solVertex; // take solution from this vertex  
  if(V_whatInType(vertex) == 3) { // if the vertex is in interior
    solVertex = vertex;
  }
  else { // if the vertex is on model bdry.
    bool noOtherVertex;
    int numEdges = V_numEdges(vertex);
    for(int i=0; i<numEdges; i++) {
      pEdge edge = V_edge(vertex,i);
      pVertex otherVertex = E_otherVertex(edge,vertex);
      if(V_whatInType(otherVertex) == 3) {
	solVertex = otherVertex;
	noOtherVertex = false;
      }
    }
    if(noOtherVertex) { // vertex is on bdry. with no interior vertex around
      cout<<"No interior node : V_info(vertex)"<<endl;
      exit(0);
    }
  }
 
  EN_getDataPtr((pEntity)solVertex,phasta_solution,(void**)&nodalSolution);

  double hmaxOther = 0.2; // user should specify this
  for (int jRow=0; jRow<3; jRow++) {
    double dot = 0.;
    for(int iDir=0; iDir<3; iDir++) {
      dot +=nodalSolution[iDir+1]*dir[jRow][iDir];
    }
    // parallel plates (i.e, w=f(y)) (only z-velocity)
    // currently velocity vector & metric is used to find the other direction
    // hmaxOther plays role in x- and y-direction (perpendicular to velocity)
    if(dot*dot < tol*tol ) // criteria to decide other hmax direction 
      if(h[jRow] > 0.2) h[jRow] = hmaxOther;
  }
}
  
