#include "phAdapt.h"


using namespace std;



extern pMeshDataId nodalhessianID;
  
// simple average over a patch surrounding the vertex    
void 
SmoothHessians(pMesh mesh) {

  pVertex v;
  VIter vIter=M_vertexIter(mesh);
  int vCount=0;
  
  // keep the vals in memory before finally setting them
  double** averageHessian;
  averageHessian = new double*[M_numVertices(mesh)];  
  
  while(v = VIter_next(vIter)) {
    
    // get each vert's surrounding nodes 
    // if they are off-bdry
    int numEdges = V_numEdges(v);
    pEdge edge;
    
    averageHessian[vCount] = new double[6];
    
    for(int iHess=0;iHess<6;iHess++) {
      averageHessian[vCount][iHess]=0.0;
    }
    // can be different from V_numEdges
    // will include the actual vertex, too
    int numSurroundingVerts=0;
    
    // exclude non-interor vertex from contributing to
    // its own average
    if(V_whatInType(v)== 3) {
      // first the "central" vertex
      double *nodalHessian;
      if(!EN_getDataPtr((pEntity)v, nodalhessianID,(void**)&nodalHessian)){
	
	cout<<"\nerror in SmoothHessians: no data attached to  vertex\n";
	V_info(v);
	exit(0);
      }
      for(int iHess=0; iHess<6; iHess++) {
	averageHessian[vCount][iHess]= averageHessian[vCount][iHess] + nodalHessian[iHess];
      }
      numSurroundingVerts++;
    }//if(V_whatInType(v)== 3) 
    
    // now for the surrounding vertices
    for(int i=0; i<numEdges ; i++){
      edge = V_edge(v,i);
      pVertex otherVertex;
      otherVertex = E_otherVertex(edge,v);
      
      // if neighbor vert is NOT on bdry
      if(V_whatInType(otherVertex)== 3){//is in interior
	
	double *nodalHessianOther;
	if(!EN_getDataPtr((pEntity)otherVertex, nodalhessianID,(void**)&nodalHessianOther)){
	  
	  cout<<"\nerror in SmoothHessians: no data attached to OTHER vertex\n";
	  V_info(otherVertex);
	  exit(0);
	}
	// add values up
	for(int iHess=0;iHess<6;iHess++) {
	  averageHessian[vCount][iHess] = averageHessian[vCount][iHess] + nodalHessianOther[iHess];
	}
	numSurroundingVerts++;
      }
    }//for( ..i<numEdges...)

    if(!numSurroundingVerts){
      cout<<"\nerror in SmoothHessians: there is a boundary vertex whose\n"
	  <<"neighbors are exclusively classsified on modelfaces/edges/vertices\n"
	  <<"and NOT in the interior\n";
      
      cout<<"For the following vertex : "<<endl;
      V_info(v);
      exit(0);
    }

//      if(!numSurroundingVerts) {   
//        double *nodalHessianSame;
//        if(!EN_getDataPtr((pEntity)v, nodalhessianID,(void**)&nodalHessianSame)){
	
//  	cout<<"\nerror in SmoothHessians: no data attached to vertex\n";
//  	V_info(v);
//  	exit(0);
//        } 
//        // add values up
//        for(int iHess=0;iHess<6;iHess++) {
//  	averageHessian[vCount][iHess] = averageHessian[vCount][iHess] + nodalHessianSame[iHess];
//        }
//        numSurroundingVerts=1;
//      }
    
    for(int iHess=0;iHess<6;iHess++) {
      averageHessian[vCount][iHess] = averageHessian[vCount][iHess]/numSurroundingVerts;
    }
    vCount++;
  }// while v = vIter
  VIter_reset(vIter); 
  
  vCount=0;
  while(v = VIter_next(vIter)) {
    // delete the old hessian data
    double* oldHessian;
    if(!EN_getDataPtr((pEntity)v, nodalhessianID,(void**)&oldHessian)){
      
      cout<<"\nerror in SmoothHessians: no data attached to OTHER vertex\n";
      V_info(v);
      exit(0);
    }
    delete [] oldHessian;
    
    EN_deleteData((pEntity)v, nodalhessianID);
    
#ifdef DEBUG
    //         cout<<"\nin SmoothHessians: going to attach\n";
    //         for(int i=0; i<6; i++){
    //             cout<<averageHessian[vCount][i]<<" ";
    //         }
    //         cout<<"\n";
#endif//DEBUG
    
    EN_attachDataPtr( (pEntity)v, nodalhessianID, (void *)
		      averageHessian[vCount] );
    
    vCount++;
  }
  VIter_delete(vIter);
}
  
