#include <math.h>
#include "phAdapt.h"

using namespace std;

// using std::cout;
// using std::endl;

void V_AnalyticHessian (pVertex v, double H[3][3], int option)
{

  for(int i=0; i<3; i++) {
    for(int j=0; j<3; j++) {
      H[i][j]=0.0;
    }
  }
  
  if(V_whatInType(v)!=3) {
    double temp_H[3][3];        
    int numSurroundingVerts=0;
    int numEdges = V_numEdges(v);

    pEdge edge;
    for(int i=0; i<numEdges ; i++) {      
      edge = V_edge(v,i);
      pVertex otherVertex;
      otherVertex = E_otherVertex(edge,v);
      
      // if neighbor vert is NOT on bdry
      if( V_whatInType(otherVertex)==3 ){//is in interior
	V_AnalyticHessian(otherVertex,temp_H,option);
	numSurroundingVerts++;
	for(int i=0; i<3; i++) {
	  for(int j=0; j<3; j++) {
	    H[i][j]+=temp_H[i][j];
	  }
	}
      }
    }//for( ..i<numEdges...)


    if(!numSurroundingVerts){
      cout<<"\nerror in V_AnalyticHessian: there is a boundary vertex whose\n"
	  <<"neighbors are exclusively classsified on model faces/edges/vertices\n"
	  <<"and NOT in the interior\n";
      
      cout<<"For the following vertex : "<<endl;
      V_info(v);
      exit(0);
    }

    for(int i=0; i<3; i++) {
      for(int j=0; j<3; j++) {
	H[i][j]=H[i][j]/numSurroundingVerts;
      }
    }

  }
  else {
    // can provide different analytic hessian 
    // to compare it with numerical one for different cases
    switch(option) {
    // case 1 not allowed as numerical hessian is used then (in sizefield.cc)
    case 2: {
      // this is generated by maple 
      // this case corresponds to 1_over_7 (turbulent) profile
      // for a straight cylindrical pipe
      double xyz[3];
      V_coord(v,xyz);
      
      double x;
      double y;
      double eps=1.0e-10;
      double t20;
      double t6;
      double ddf_xy;
      double t7;
      double t4;
      double t3;
      double t2;
      double t1;
      double t16;
      double t17;
      double t15;
      double t14;
      double t12;
      double t11;
      double t10;
      double t9;
      double t8;
      double t29;
      double t25;
      double t24;
      double t21;
      double t32;
      double t30;
      double t31;
      double t37;
      double t36;
      double t40;
      double t42;
      double t41;
      double t45;
      double t48;
      double t57;
      double t74;
      double t56;
      x = xyz[0];
      y = xyz[1];
      t1 = x * x;
      t2 = y * y;
      t3 = t1 + t2;
      t4 = sqrt(t3);
      t6 = 0.250e2 + eps - 0.250e2 * t4;
      t7 = pow(t6, -0.20e1);
      t8 = 0.1e1 +eps - t4;
      t9 = pow(t8, -0.2857142857e0);
      t10 = t7 + t9;
      t11 = t10 * t10;
      t12 = sqrt(t10);
      t14 = 0.10e1 / t12 / t11;
      t15 = pow(t6, -0.30e1);
      t16 = 0.10e1 / t4;
      t17 = t15 * t16;
      t20 = pow(t8, -0.1285714286e1);
      t21 = t20 * t16;
      t24 = 0.5000e2 * t17 * x + 0.2857142857e0 * t21 * x;
      t25 = t24 * t24;
      t29 = 0.10e1 / t12 / t10;
      t30 = pow(t6, -0.40e1);
      t31 = 0.10e1 / t3;
      t32 = t30 * t31;
      t36 = 0.10e1 / t4 / t3;
      t37 = t15 * t36;
      t40 = 0.5000e2 * t17;
      t41 = pow(t8, -0.2285714286e1);
      t42 = t41 * t31;
      t45 = t20 * t36;
      t48 = 0.2857142857e0 * t21;
      t56 = 0.5000e2 * t17 * y + 0.2857142857e0 * t21 * y;
      t57 = t56 * t56;
      t74 = y * x;
      ddf_xy = 0.3e1 / 0.4e1 * t14 * t24 * t56 - t29 * (0.37500000e4 * t32 * t74 - 0.5000e2 * t37 * t74 + 0.3673469388e0 * t42 * t74 - 0.2857142857e0 * t45 * t74) / 0.2e1;
      H[0][0] = 0.3e1 / 0.4e1 * t14 * t25 - t29 * (0.37500000e4 * t32 * t1 - 0.5000e2 * t37 * t1 + t40 + 0.3673469388e0 * t42 * t1 - 0.2857142857e0 * t45 * t1 + t48) / 0.2e1;
      H[0][1] = ddf_xy;
      H[0][2] = 0.0e0;
      H[1][0] = ddf_xy;
      H[1][1] = 0.3e1 / 0.4e1 * t14 * t57 - t29 * (0.37500000e4 * t32 * t2 - 0.5000e2 * t37 * t2 + t40 + 0.3673469388e0 * t42 * t2 - 0.2857142857e0 * t45 * t2 + t48) / 0.2e1;
      H[1][2] = 0.0e0;
      H[2][0] = 0.0e0;
      H[2][1] = 0.0e0;
      H[2][2] = 0.0e0;
    }
    break;
    default : {
      cout<<"\nSpecify a correct option (V_AnalyticHessian)"<<endl;
      exit(-1); 
    }
    break;
    }
  }
}
