#include <iostream>
#include <fstream>
#include <string>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifndef CYGWIN
  #include <unistd.h>
#endif

#ifdef CYGWIN
  #include <direct.h>
  #define chdir _chdir
#endif

using namespace std;


#include "SimParasolidKrnl.h"
#include "SimAdvMeshing.h"

#include "phAdapt.h"
#include "phReadWrite.h"
#include "attachData.h"




// avoid elements at (bdry.) with no interior nodes/dofs
extern void fix4NodesOnSurface(pMesh mesh);



int CBcounter=0;
int CBinvalidRegionCount=0;


pMeshDataId phasta_solution;
pMeshDataId errorIndicatorID;
pMeshDataId nodalgradientID;
pMeshDataId nodalhessianID;
pMeshDataId modes; 


int phAdapt(//time step
	    int lstep,
	    // strategy is to specify 
	    // how to do adaptation (i.e., size-field or tag driven)
	    // 1-2 : size-field driven (for anisotropic) 
	    // 3-4 : tag driven (for isotropic)
	    // 5-6 : size-field driven (for isotropic)
	    // < 0 sets a manual mesh size-field
	    int strategy,
	    // factor is the constant appearing in the error expression
	    // for tag driven it is used to define threshold for refinement
	    // for size-field driven it is used to define the error tolerance
	    double factor,
	    // number of solution variables (5 for incompressible)
	    int ndof,
	    // number of variables for error indicators (EI)
	    // (e.g., 5 for ybar & 10 for residual-based)
	    int nvar,
	    // the maximal mesh edge length allowed in mesh size 
	    double hmax,
	    // the minimal mesh edge length allowed in mesh size
	    double hmin,
	    // option is used to decide how to compute the error value
	    // provides different choices like analytic hessian, manual size-field etc.
	    // for isotropic (ta or size-field driven :
	    // use 3 EI for flow problem or use 1 EI for scalar problem
	    int option)
{  
  // polynomial order
  int poly = 1;   
  // number of nodes (as considering only linear basis)
  int nshg;       

  char model_file[28];
  char mesh_file[28];
  char solution_file[28];
  char error_indicator_file[28];

  sprintf(model_file,"geom.xmt_txt");  
  sprintf(mesh_file,"geom.sms");  

  MS_init();
  Sim_readLicenseFile(0);

  SimParasolid_start(1);


  pParasolidNativeModel nModel;
  pGModel model;


  // a file containing your feature keys,
  // or you can use individual calls to Sim_registerKey()
  // NOTE: Sim_readLicenseFile() is for internal testing only.  For
  // a release product, use Sim_registerKey()


  nModel = ParasolidNM_createFromFile("geom.xmt_txt",0);
  model = GM_createFromNativeModel(nModel);



  pMesh mesh=M_new(0,model);
  mesh = M_load(mesh_file,model);


  printf("\n-- Before Adaptation...\n");
  printf(" Total # of elements: %d\n", M_numRegions(mesh));
  printf(" Total # of vertices: %d\n", M_numVertices(mesh));

  cout<<endl;
  cout<<"MeshSim   Library build ID : "<<SimMeshing_buildID()<<endl;
  cout<<"MeshTools Library build ID : "<<SimMeshTools_buildID()<<endl;
  cout<<endl;

  nshg=M_numVertices(mesh);
  phasta_solution = MD_newMeshDataId("restart solution");
  errorIndicatorID = MD_newMeshDataId("error indicator");
  modes = MD_newMeshDataId("number of modes");// required for higher order


  pMSAdapt simAdapter; 

  // if strategy<0 (command line argument as of now) 
  // then sets a manual size field  
  switch(strategy) {
  case 1 :
  case 2 : { //anisotropic adaptation
    cout<<"\nStrategy chosen for ANISOTROPIC adaptation : size-field driven"<<endl;
    
    sprintf(solution_file,"restart.%i.0",lstep);

    char error_tag[28];
    if(option==1) {
      if(strategy == 1) {
	cout<<"\nUsing ybar to compute hessians...\n"<<endl;
	sprintf(error_tag,"ybar");
	sprintf(error_indicator_file,"ybar.%i.0",lstep);
      }
      else if (strategy == 2) {
	cout<<"\nUsing numerical/computed hessians (i.e, from phasta)...\n"<<endl;
	sprintf(error_tag,"hessains");
	sprintf(error_indicator_file,"restart.%i.0",lstep);
      }
      
      cout<<"\n Reading files:"<<endl;
      cout<<" ..."<<solution_file<<" (for \"solution\")"<<endl;
      cout<<" ..."<<error_indicator_file<<" (for \""<<error_tag<<"\")"<<endl;
      cout<<endl;
    }//option ==1
    else
      cout<<"\nUsing analytic hessians...\n"<<endl;

    // 2nd arg. (1 : size-field driven adaptation)
    simAdapter = MSA_new(mesh,1);

    // invoke solution transfer
//JM26/9
    MSA_setCallback(simAdapter, phastaTransfer,9,NULL);// 8 for regions, 1 for vertices

    if(option==1 && strategy==1) {
      // attaching the solution to the original mesh
      double *sol;
      readArrayFromFile(solution_file,"solution",sol);
      attachArray(sol,mesh,phasta_solution,ndof,poly);
      delete [] sol;

      // read ybar (and compute/use hessians of ybar) 
      double *error_indicator;
//JMtmp 9/25
      readArrayFromFile(error_indicator_file,error_tag,error_indicator);
      attachArray(error_indicator,mesh,errorIndicatorID,nvar,poly);
      delete [] error_indicator;
      
      // calculating hessians for ybar field
      // first reconstruct gradients and then the hessians 
      // also deals with boundary issues &
      // applies smoothing procedure for hessians
      // (simple average : arithmetic mean)
//JMtmp 9/25
      hessiansFromSolution(mesh,lstep);
    }
    else if (strategy == 2) { // cannont use analytic hessian in this case
      // option : choice of variable (whose hessians are taken)
      // use the hessians computed from phasta
      double *hessiansFromRestart;
      readArrayFromFile(error_indicator_file,error_tag,hessiansFromRestart);
      double *hessians = new double[nshg*6];
      V_getHessians(hessiansFromRestart,mesh,nvar,option,hessians);
      delete [] hessiansFromRestart;
      attachArray(hessians,mesh,nodalhessianID,6,poly);
      delete [] hessians;

      // once option is used in V_getHessians 
      // it is changed so that we don't use analytic hessian (in sizefield.cc)
      option = 1;

      // simple average over a patch surrounding the vertex 
      SmoothHessians(mesh);      
    }
    
    // compute mesh size-field using hessian strategy (anisotropic adaptation)
    // and set it at each vertex
//JMtmp 9/25
    setSizeFieldUsingHessians(mesh,simAdapter,factor,hmax,hmin,option);
  }
  break;
  case 3:
  case 4: { // isotropic adaptation (tag driven)
    cout<<"\nStrategy chosen for ISOTROPIC adaptation : tag driven"<<endl;
    sprintf(solution_file,"restart.%i.0",lstep);

    char error_tag[28];
    if(strategy == 3) {
      sprintf(error_tag,"error");
      sprintf(error_indicator_file,"error.%i.0",lstep);
    }
    else if (strategy == 4) {
      sprintf(error_tag,"error");
      sprintf(error_indicator_file,"restart.%i.0",lstep);
    }

    cout<<"\n Reading files:"<<endl;
    cout<<" ..."<<solution_file<<" (for \"solution\")"<<endl;
    cout<<" ..."<<error_indicator_file<<" (for \""<<error_tag<<"\")"<<endl;
    cout<<endl;
    
    // 2nd arg. (0 : tag driven & 1 : size-field driven adaptation)
    simAdapter = MSA_new(mesh,0);

    // invoke solution transfer
    MSA_setCallback(simAdapter, phastaTransfer,9,NULL);// 8 for regions, 1 for vertices

    // attaching the solution to the original coarse mesh
    double *sol;
    readArrayFromFile(solution_file,"solution",sol);
    attachArray(sol,mesh,phasta_solution,ndof,poly);
    delete [] sol;
    
    // read residual based error indicators (10 quantities)
    double *error_indicator;
    readArrayFromFile(error_indicator_file,error_tag,error_indicator);
    attachArray(error_indicator,mesh,errorIndicatorID,nvar,poly);

    // for(int iError=0; iError<nshg; iError++)
    // cout<<"Error "<<iError<<" : "<<error_indicator[iError*10]<<endl;

    delete [] error_indicator;

    // tag the entities to be refinement (for isotropic refinement)
    // factor is used to evaluate the threshold for refinement
    // as of now do not use hmin and hmax
    tagEntitiesForRefinement(mesh,simAdapter,factor,hmax,hmin,option);
  }
  break;
  case 5:
  case 6: { //isotropic adaptation (size-field driven)
    cout<<"\nStrategy chosen for ISOTROPIC adaptation : size-field driven"<<endl;

    sprintf(solution_file,"restart.%i.0",lstep);

    char error_tag[28];
    if(strategy == 5) {
      sprintf(error_tag,"error");
      sprintf(error_indicator_file,"error.%i.0",lstep);
    }
    else if (strategy == 6) {
      sprintf(error_tag,"error");
      sprintf(error_indicator_file,"restart.%i.0",lstep);
    }

    cout<<"\n Reading files:"<<endl;
    cout<<" ..."<<solution_file<<" (for \"solution\")"<<endl;
    cout<<" ..."<<error_indicator_file<<" (for \""<<error_tag<<"\")"<<endl;
    cout<<endl;

    // 2nd arg. (0 : tag driven & 1 : size-field driven adaptation)
    simAdapter = MSA_new(mesh,1);

    // invoke solution transfer
    MSA_setCallback(simAdapter, phastaTransfer,9,NULL);// 8 for regions, 1 for vertices


    // attaching the solution to the original coarse mesh
    double *sol;
    readArrayFromFile(solution_file,"solution",sol);
    attachArray(sol,mesh,phasta_solution,ndof,poly);
    delete [] sol;
    
    // read residual based error indicators (10 quantities)
    double *error_indicator;
    readArrayFromFile(error_indicator_file,error_tag,error_indicator);
    attachArray(error_indicator,mesh,errorIndicatorID,nvar,poly);

    // for(int iError=0; iError<nshg; iError++)
    // cout<<"Error "<<iError<<" : "<<error_indicator[iError*10]<<endl;

    delete [] error_indicator;

    // currently we are smoothing the residual based part 
    // of the set of all EIs given through phasta (which might be already smoothed)
    // the quantity that is smoothed (arithmetic mean) and attached:
    // res_u^2 + res_v^2 + res_w^2
    // (the first three components from the set of ten)
    // the ID used to attach the smoothed EIs is "errorIndicatorID"
    // data previously attached under that ID is deleted
    SmoothErrorIndicators(mesh,option);

    writeSmoothEIs(mesh);

    // compute mesh size-field using residues (isotropic adaptation)
    // and set it at each vertex
    setIsotropicSizeField(mesh,simAdapter,factor,hmax,hmin,option); 
  }
  break;
  default : {
    if(strategy<0) {
      cout<<"\nStrategy chosen for adaptation : manual size-field"<<endl;

      // 2nd arg. (1 : size-field driven adaptation)
      simAdapter = MSA_new(mesh,1);
      
      // set mesh size-field manually (anisotropic adaptation) at each vertex
      // strategy provides different choices
      setManualSizeField(mesh,simAdapter,strategy);
    }
    else {
      cout<<"\nSpecify a correct (adaptation) strategy (adapt.cc)"<<endl;
      exit(-1);
    }
  }
  break;
  }
  
  // data clean up
//JMtmp 9/25
  if(strategy==1 || strategy==3 || strategy==4)
      cleanAttachedData(mesh,errorIndicatorID,0);
  else if (strategy==5 || strategy==6)
      cleanAttachedData(mesh,errorIndicatorID,0,0);


  // use mesh callback function to do solution transfer
  // MSA_setCallback(simAdapter, phastaTransfer);

  // adaptation

//JMtmp 9/25
  MSA_adapt(simAdapter);

  // takes case of bad brdy. elements (elements with no interior nodes)
  fix4NodesOnSurface(mesh);
  cout<<"fixing solution transfer\n";
  fix4SolutionTransfer(mesh);



  printf("-- Adaptation Done...\n");
  printf(" Total # of elements: %d\n", M_numRegions(mesh));
  printf(" Total # of vertices: %d\n\n", M_numVertices(mesh));
  printf(" Total # of sol. callbacks: %d\n\n",CBcounter);
  printf(" Total # of invalid regions in  callbacks: %d\n\n",CBinvalidRegionCount);

  // create a new directory
  // put the adapted mesh and solution in this new directory
  char sss[100];
  sprintf(sss,"mkdir %d",lstep);
  system(sss);
  //sprintf(sss,"cd %d",lstep);
  sprintf(sss,"%d",lstep);
  //system(sss);
  chdir(sss);

  printf("\n Writing out the mesh...\n\n");
  M_write(mesh,mesh_file,5);

  // we need this when we employ solution transfer
  // writing out the solution into a restart file
//JM 26/9
  printf(" Writing out solution...\n\n");
  char output_file[28];
  int nshg_fine = M_numVertices(mesh);
  sprintf(output_file,"restart.%d.1",lstep);
  
  double* solution;
  getAttachedArray(solution,mesh,phasta_solution,ndof,poly);
  


  writeArrayToFile(output_file,"solution","binary","write",
                   nshg_fine,ndof,lstep,solution);

  delete [] solution;



  // 3rd arg. (en_type) : 0 for vertex, 1 for edge etc.
  if(strategy > 0){
      
      cleanAttachedData(mesh,phasta_solution,0);
  }


  MD_deleteMeshDataId(phasta_solution);
  MD_deleteMeshDataId(errorIndicatorID);
  MD_deleteMeshDataId(modes);

  MSA_delete(simAdapter);
  M_release(mesh);



  GM_release(model);
  NM_release(nModel);
  SimParasolid_stop(1);  


  MS_exit();    


  return 1;
} 

void check(pMesh mesh)
{
  int nv=0, ne=0, nf=0, nr=0, unknown=0;
  
  VIter vit=M_vertexIter(mesh);
  pVertex vertex;
  while( vertex=VIter_next(vit) ) {
    switch( V_whatInType(vertex) ) {
    case Tvertex: nv++; break;
    case Tedge: ne++; break;
    case Tface: nf++; break;
    case Tregion: nr++; break;
    default:
      unknown++;
    }
  }
  
  printf("# of vertices on gv: %d\n",nv);
  printf("# of vertices on ge: %d\n",ne);
  printf("# of vertices on gf: %d\n",nf);
  printf("# of vertices on gr: %d\n",nr);
  printf("# of vertices of unknown classification: %d\n",unknown);
}

