#include <stdio.h>
#include <stdlib.h>
#include "MeshSim.h"
#include "MeshSimAdapt.h"
#include "phAdapt.h"



// implementation of the callback fctn
extern int CBcounter;
extern int CBinvalidRegionCount;
extern pMeshDataId phasta_solution;
extern pMeshDataId modes;
extern int ndof;
//for measuring procsize

// arguments are :
// mesh mod. type, look at file MeshSimAdapt.h
// data containing mesh entities created, deleted or reshaped
// userData used to set a pointer to be passed in callback function
void
phastaTransfer( MeshModType mtype, pMeshChanges mco, void *userData)  {

    CBcounter++;

    pPList fresh;
    pPList old;
    pRegion  parent_region=0;
    pRegion  rgn;
    pEntity ent;
    int numFoundOldSolution;
    int count = 0;
    int counter = 0;

    pPList newVertices;

    pPList verticesO;
    pPList verticesN;

    pVertex v;
    pVertex deleteVertex;
    int i,j,k,deletedVerts;

    double xyz[3], xietazeta[3];
    double* nodalSolution;
    double coords[3];
//     FILE  *logfile; 
    char stringName[255];
    double* averageSolution;
    double* avSol;
    double* solution;
    
//     sprintf(stringName, "adaptLog.%d",PMU_rank+1);
    //logfile=fopen(stringName,"a");
//    printf("\n entering callback fct\n");

    pPList freshEnts = MCO_createdEntities(mco); 
    pPList oldEnts  = MCO_deletedEntities(mco);

    int oldEntsSize = PList_size(oldEnts);
    int freshEntsSize = PList_size(freshEnts);

    // this is done as callback function is invoked for
    // some mesh mod. operations which are not successful
    // this must be fixed with newer versions of Simmetrix libs.
    if(!freshEntsSize) {
      PList_delete(freshEnts);
      PList_delete(oldEnts);

      // do not count this call
      CBcounter--;

      return;
    }

    // old - holds old regions (to be deleted)
    old = PList_new();
    // freash - holds new regions (created)
    fresh = PList_new();
    // oldVertices - hold old vertices (to be deleted)
    pPList oldVertices = PList_new();
    
    for(i=0; i<oldEntsSize; i++){
        ent = (pEntity)PList_item(oldEnts,i);
        int en_type = EN_type(ent);
        if(en_type==1) {
            PList_append(oldVertices,ent);
        }
        else if(en_type==3) {
            PList_append(old,ent);
        }
    }

    for(i=0; i<freshEntsSize; i++){
        ent = (pEntity)PList_item(freshEnts,i);
        int en_type = EN_type(ent);
        if(en_type==3) {
            PList_append(fresh,ent);
        }
    }

    PList_delete(freshEnts);
    PList_delete(oldEnts);

    numFoundOldSolution = 0;
    
    ////////////////////////////////////////////////////////////////////////////////////////
    // more than 1 region in old
    ////////////////////////////////////////////////////////////////////////////////////////
    if ( PList_size( old ) > 1 ) {
        
        count = PList_size( old );
        

//        printf("\nsize of old ent list %d  \n",count);

/*          printf("\nMesh mod type is %d and sizeof(old ent list) >1\n",mtype); */
//        PList_printx(old); 
//           PList_printx(fresh); 
        // get all vertices that have some solution on them and build average
        // set this average to all NEW vertices of the NEW regions
        averageSolution= (double *)malloc(sizeof(double)*ndof );
        for( i =0; i <ndof ; i++ ){
        averageSolution[i]=0.0;
        }

        
        for( i =0; i < count ; i++ ){
             
            verticesO = R_vertices((pRegion)PList_item(old , i ),1);

            

/*              printf("\n counter is  %d\n",counter); */
/*              printf("\ngot %d old vertices of region %d\n",PList_size(verticesO),i); */

            for( j =0; j < PList_size(verticesO) ; j++ ){

                v = (pVertex)PList_item(verticesO , j );
//                printf("\n got ONE old vertex num %d of region %d\n",j,i );


                if( EN_getDataPtr( ( pEntity )v, phasta_solution, (void
                                                                    **)&nodalSolution) != 0){
                    for( k =0; k <ndof ; k++ ){
                        averageSolution[k] += nodalSolution[k];
                    }
                    numFoundOldSolution++;
//                    printf("\n found solution on old vertex \n" );
                }
            }

	    PList_delete(verticesO);

            counter++;
        }
        if(numFoundOldSolution != 0){
            for( i =0; i <ndof ; i++ ){
                averageSolution[i] = averageSolution[i]/numFoundOldSolution;
            }
        }
        else{
            printf("\n all vertices in old region list have NO solution \n" );
            exit(1);
        }
        // assign the average to ALL new vertices
        for( i =0; i < PList_size( fresh ) ; i++ ){

//            printf("\n got vertices of a new region \n" );

            verticesN = R_vertices((pRegion)PList_item(fresh , i ),1);



            for( j =0; j < PList_size(verticesN  ) ; j++ ){

                v =  (pVertex)PList_item(verticesN , j );


                if(! EN_getDataPtr( ( pEntity )v, phasta_solution, (void
                                                                    **)&nodalSolution)){
                
                    // provide memory for new solution
		    // avSol= (double *)malloc(sizeof(double)*ndof );
		    avSol = new double[ndof];
                    for( i =0; i <ndof ; i++ ){
                        avSol[i]=averageSolution[i];
                    }
//                    printf("\n found vertex in new regions that got assigned a sol \n" );
                    EN_attachDataPtr( (pEntity)v, phasta_solution, (void *)
                                      avSol ); 
                }
            }

	    PList_delete(verticesN);
        }
        free(averageSolution);


        
    }// PList_size(old) > 1 
    ////////////////////////////////////////////////////////////////////////////////////////
    // exactly 1 region in old list ==> it is  a split operation
    ////////////////////////////////////////////////////////////////////////////////////////
    else if( PList_size( old ) == 1){
//        printf("\n  1 region in old, split type is %d\n",mtype );
        
        parent_region = (pRegion)PList_item( old, 0 );
        
        // split operation
        // retrieve new vertices
        // get all vertices on old and new element(s)
        verticesO = R_vertices(parent_region,1);
        
        newVertices = PList_new();

        for( i =0; i < PList_size( fresh ) ; i++ ){

            verticesN = R_vertices((pRegion)PList_item(fresh , i ),1);

            for( j =0; j < PList_size(verticesN  ) ; j++ ){
                
                if(!PList_inList(newVertices,PList_item(verticesN , j ))){
                     PList_append(newVertices,PList_item(verticesN , j ));
                }
            }

	    PList_delete(verticesN);
        }         
        // find the vertices which are not in the old list (-> a new vertex)
        for( i =0; i < PList_size(newVertices  ) ; i++ ){
            
            v =  (pVertex)PList_item(newVertices , i );
            V_coord(v,xyz);
             
            if(! PList_inList(verticesO,v)){
                
                //if there is not a solution attached already
                if(! EN_getDataPtr( ( pEntity )v, phasta_solution, (void
                                                                    **)&nodalSolution)){


                    // do a solution interpolation for that vertex
                    // calculate the parametric co-ordinates of the current vertex in the
                    // space of the parent region. ( also use these to make sure that the
                    // vertex lies inside the given region )
                    // inversemap implemented in InverseMap.cc
//                      if(! R_isValidElement(parent_region)){

//                          cout<<"Warning in callback: found an invalid element by using R_isValidElement\n";
//                          exit(0);
//                      }
                    if ( !inverseMap( parent_region, xyz, xietazeta ) ) {	
                        printf("Warning in CB: Could not locate point inside the parent region: \n");
                        //R_info(parent_region);
                        //V_info(v);
                        CBinvalidRegionCount++;
                        //exit( -1 );
                    }
                    else {	// inverse Map IS valid 
                                // be invalid
                        // interpolate the solution from the original region onto the finer one 
                        // use appropriate shapeFuns             
                        
                        // check if all verts of parent_region have data
                        // attached
                        int valid=1;
                        double* foo;
                        pPList parentVerts = R_vertices(parent_region,1);
                        for(int nu=0;nu< PList_size( parentVerts ) ; nu++ ){

                            pVertex vx = (pVertex)PList_item(parentVerts , nu );

                            if(! EN_getDataPtr( ( pEntity )vx, phasta_solution, (void
                                                                    **)&foo)){
                                
                                valid=0;
                            }
                        }
                        if(valid){
                            solution = InterpolateSolution( parent_region,
                                                            xietazeta,
                                                            ndof,
                                                            modes);
                            
                            // attach it to the vertex (if not done yet)
                            // otherwise do nothing
                            EN_attachDataPtr( (pEntity)v, phasta_solution, (void *)
                                              solution );
                            
//                    printf("\n info: assigning a solution val to a new vertex: \n" );
/*                      V_coord( v ,coords ); */
/*                      for(j=0; j<3; j++){ */
/*                          //fprintf(logfile,"%f\n",coords[j]); */
/*                      } */
                            //fprintf(logfile,"\n");
                        }//if(valid
                    }//else
                }//if(! EN_getDataPtr(  
                else{
                    // a solution already exists at that vertex: do nothing
                }
                
            }// if(! PList_inList(v
        }//for(...
       

        PList_delete(verticesO);
        PList_delete(newVertices);
    }
    else{//  size of old regions list =0
//        printf("\n  in callback.c: size of old regions list =0 \n" );
    }


    PList_delete(fresh);
    PList_delete(old);


    for( i =0; i < PList_size(oldVertices) ; i++ ){
        ent = (pEntity)PList_item(oldVertices,i);
        double* data;
        if(!EN_getDataPtr( ( pEntity )ent, phasta_solution, (void
							     **)&data)) {
	  printf("\nError in callback.cc : solution NOT attached to an old vertex\n");
	  exit(1);
	}
        
        delete [] data;
	EN_deleteData((pEntity)ent,phasta_solution);
    }
    PList_delete(oldVertices);
    
/*      printf("\n  in callback.c:exit callback function \n" ); */
/*      printf("[%d] exiting Callback.cc procSize (MY) : %d \n",PMU_rank(),myProcSize()); */
}

