#include <iostream>
#include "phAdapt.h"


using namespace std;


extern pMeshDataId nodalgradientID;

// reconstruct the element hessian : 6-component (symmetric)
void
elementHessian(pRegion region, double* elemHessian)
{
  pVertex v;
  double* nodalGradient;
  
  // build the linear system
  double matrix[16];
  buildSystem(region,matrix);
  
  double  fieldVectorX[4];
  double fieldVectorY[4];
  double fieldVectorZ[4];
  int four = 4;
  int indx[4];
  double fnumber;
  
  pPList regionVerts;
  regionVerts = R_vertices(region, 0);
  
  // get the field vals
  for (int i=0; i<PList_size(regionVerts); i++) {
    v = (pVertex)PList_item(regionVerts,i);
    if(EN_getDataPtr((pEntity)v, nodalgradientID,
		     (void**)&nodalGradient)==NULL){
      cout<<"\nerror in elementHessian: no data attached to vertex\n";
      exit(0);
    }
    fieldVectorX[i]= nodalGradient[0];
    fieldVectorY[i]= nodalGradient[1];
    fieldVectorZ[i]= nodalGradient[2];
  }
  PList_delete(regionVerts);    
  
  ludcmp(matrix , &four, &four, indx, &fnumber);
  
  // fieldVector is going to be overridden and will
  // contain the solution
  lubksb(matrix , &four, &four, indx, fieldVectorX );
  lubksb(matrix , &four, &four, indx, fieldVectorY );
  lubksb(matrix , &four, &four, indx, fieldVectorZ );

  // each poly is 
  // a0 + a1*x + a2*y + a3*z  
  elemHessian[0] = fieldVectorX[1];// xx
  elemHessian[1] = fieldVectorX[2];// xy
  elemHessian[2] = fieldVectorX[3];// xz
  elemHessian[3] = fieldVectorY[2];// yy
  elemHessian[4] = fieldVectorY[3];// yz
  elemHessian[5] = fieldVectorZ[3];// zz
  
#if  ( defined  DEBUG )
  //     for(int i=0; i<6; i++) {
  //         cout<<"\nhessian["<<i<<"] "<<elemHessian[i]<<"\n";
  //     }
#endif
 
}
   
