#include "phAdapt.h"



extern pMeshDataId nodalgradientID;
extern pMeshDataId nodalhessianID;
      
// recover hessian using a patch (from gradients)
// hessian returned : 6-component (symmetric)
// u_xx, u_xy, u_xz, u_yy, u_yz, u_zz
void
hessiansFromPatch(pMesh mesh)
{
  VIter vIter;
  vIter = M_vertexIter(mesh);
  pVertex vertex;
  
  double elemHessian[6];
  double patchVolume;
  
  // loop over vertices and get patch of elements
  while(vertex = VIter_next(vIter)) {
    pPList elementPatch;
    elementPatch = V_regions(vertex);
    patchVolume=0;

    double* patchHessian = new double[6];
    for(int i=0; i<6; i++) {
      patchHessian[i]=0.0;
    }
    
    // loop over elements and recover hessian at vertex
    for (int i=0; i < PList_size(elementPatch); i++) {
      patchVolume+=R_volume((pRegion) PList_item(elementPatch,i));

      // compute element hessian (each component) to recover hessian at vertex
      elementHessian( (pRegion)PList_item(elementPatch,i), elemHessian);
      
      // use element hessian (each component) to recover hessian at vertex
      for(int j=0; j<6; j++) {
	patchHessian[j]= patchHessian[j] + 
	  elemHessian[j]* R_volume((pRegion) PList_item(elementPatch,i));
      }
      
    }//loop over element patch ends
    
    for(int i=0; i<6; i++) {
      patchHessian[i]= patchHessian[i] / patchVolume ;
    }

    // attach the recovered hessian to the vertex    
    EN_attachDataPtr( (pEntity)vertex, nodalhessianID, (void *)
		      patchHessian );

    PList_delete(elementPatch);
  }//while (vertex = VIter_next(vIter))      
  VIter_delete(vIter);

#if  ( defined  DEBUG )
  //    writeRestartHessians(mesh);
#endif//DEBUG      
}

