#include "phAdapt.h"
#include <stdlib.h>
#include <stdio.h>


// global ndof is the ttal number of solution vars to be transferred via
// callback
int ndof;
int main(int argc,char* argv[])
{
  if( argc != 9 ) {
    printf("usage : %s\n",argv[0]);
    printf("        stepNumber selectedStrategy numberSolutionVariables numberErrorVariables reductionRatio hmax/MaxCoarseFactor hmin/MaxRefineFactor option\n\n");
    printf("        - modelname is always geom.xmt_txt\n");
    printf("        - input mesh is geom.sms\n");
    printf("        - adapted mesh is geom.sms, but in another directory\n");
    printf("\n");
    printf("        - the Error Indicators (EI) are read from:\n");
    printf("          ybar.stepNumber.0 or error.stepNumber.1\n");
    printf("          (depending upon selectedStrategy)\n");
    printf("        - solution is read from restart.sn.0\n");
    printf("        - transfered solution is associated with new mesh\n");
    printf("\n");
    printf("        - selectedStrategy : how to do adaptation\n");
    printf("          (i.e., 1-2 for anisotropic & size-field driven)\n");
    printf("          (i.e., 3-4 for isotropic & tag driven)\n");
    printf("          (i.e., 5-6 for isotropic & size-field driven)\n");
    printf("        - numberSolutionVariables : number of vars. for solution transfer\n");
    printf("          (5:for P,v,T & 6:for additional scalar)\n");
    printf("        - numberErrorVariables : number of vars. for error indicators\n");
    printf("          (5:for ybar & 10:for residual based errors)\n");
    printf("        - reductionRatio : constant in error indicator\n");
    printf("          (for isotropic defines threshold for refinement)\n");
    printf("          (for anistropic defines error tolerance)\n");
    printf("        - option : how to compute error value \n");
    printf("          (0 : use 1 scalar (like temp.) EI for scalar problems) \n");
    printf("          (1 : use 3 momentum EI for flow problems) \n\n");
    printf("        - hmax & hmin : used for anisotropic adaptation\n");
    printf("          max. and min. edge length allowed, respectively\n");
    printf("        - MaxCoarseFactor & MaxRefineFactor : used for size-field driven isotropic adaptation\n");
    printf("          MaxCoarseFactor (>1.) : h^(new)/h^(old) (how much coarsening)\n");
    printf("          MaxRefineFactor (>1.) : h^(old)/h^(new) (how much refinement)\n\n");
    exit(0);
  }
  
  // step number, i.e., time step number in phasta filename
  int sn=atoi(argv[1]);
  // strategy is to specify how to do adaptation 
  // (i.e., size-field or tag driven)
  // 1-2 : size-field driven (for anisotropic) 
  // 3-4 : tag driven (for isotropic)
  // 5-6 : size-field driven (for isotropic)
  // < 0 sets a manual mesh size-field
  int strategy=atoi(argv[2]);
  // number of solution variables to be transferred
  ndof = atoi(argv[3]);

  // number of variables for error indicators (EI)
  // (e.g., 5 for ybar & 10 for residual-based)
  int nvar=atoi(argv[4]);
  // factor is the constant appearing in error indicator/estimator
  // for isotropic it is used to define the threshold for refinement
  // for anisotropic it is used to define error tolerance (in a way)
  double ratio=atof(argv[5]);
  // max. allowable edge length
  // needed in case hessian is singular and eigenvalues are zero
  double hmax=atof(argv[6]);
  // min. edge length that user wants, i.e., limit on refinement
  // needed for discontinuous solutions 
  // (esp. for shocks when hessians are not reliable)
  double hmin=atof(argv[7]);
  // option is to decide how to compute error value
  // 0 : use 1 scalar (like temp.) EI for scalar problems
  // 1 : use 3 momentum EI for flow problems
  int option=atoi(argv[8]);

  // do not need this parameter when using meshSim
  // (i.e., specified in scorec meshAdapt code)
  // maximal allowed iterations allowed in a mesh adaptation
  // int maxIter=atoi(argv[8]);

  Sim_logOn("SIM.log");
  // this is the right value for spipe example
  // `ratio' : not aspect ratio (ratio to reduce mean error)
  phAdapt(sn,strategy,ratio,ndof,nvar,hmax,hmin,option);

  Sim_logOff();

  return 1;
}
