#ifndef _PHADAPT_H_
#define _PHADAPT_H_
#include <iostream>
#include <math.h>
#include <stdlib.h>



#include "MeshSimInternal_phAdapt.h" 
#include "MeshSimAdapt.h"
#include "SimMeshTools.h"

#define ABS(x) ((x) < 0 ? -(x) : (x))
#define MAX(x,y) ((x)<(y) ? (y) : (x))


#ifdef WIN32
void  bzero(void* ptr, size_t sz); 
#endif

  
  struct Hessian {
    double h[3];
  double dir[3][3];
  };
  typedef struct Hessian Hessian;  

  // sim call back  
  // typedef int (*MSA_CallbackFunc)(pPList , pPList, int);
 
  // main routine to control mesh adaptation, solution transfer and communication with PHASTA
  int phAdapt(int,int,double,int,int,double,double,int option); 
  
  // piecewise linear mesh size field definition in terms of Hessian
  void setSizeFieldUsingHessians(pMesh,pMSAdapt,double,double,double,int option=1);
  void setIsotropicSizeField(pMesh,pMSAdapt,double,double,double,int option = 0);
  void setManualSizeField(pMesh mesh, pMSAdapt simAdapter, int comp);

  // for tag driven refinement (i.e., isotropic)
  void tagEntitiesForRefinement(pMesh,pMSAdapt,double,double,double,int option = 0);
  int applyMarkingStrategy(pMesh,pMSAdapt,double,double,double,
			   double &,double &,double &,double &,int option);
  double getErrorThreshold(pMesh,double,double &,double &,double &,int option);
  double getErrorValue(double *nodalValues,int option);

  // get interpolation error values from hessians  
  double maxLocalError(pVertex vertex, double H[3][3]);
  double E_error(pEdge edge, double H[3][3]);

  // get hessians computed from phasta
  void V_getHessians(double*,pMesh,int,int,double*);


  // for doing hessians computation outside phasta
  void V_Hessian(pVertex v, double T[3][3]);
  void V_AnalyticHessian(pVertex v, double T[3][3], int option);
  void buildSystem(pRegion region, double* eltMatrix);
  void elementGradient(pRegion region, double* elemGradient);
  void gradientsFromPatch(pMesh mesh);
  void elementHessian(pRegion region, double* elemHessian);
  void hessiansFromSolution(pMesh mesh,int stepNumber);
  void hessiansFromPatch(pMesh mesh);
  void ModifyHessiansAtBdry(pMesh mesh);
  void writeRestartHessians(pMesh mesh );
  void SmoothHessians(pMesh mesh);
  void SmoothErrorIndicators(pMesh mesh,int option);
  void writeSmoothEIs(pMesh mesh);

  void ModifyMetric(pVertex vertex, double dir[3][3], double* h);




  // for viewing results in medit 
  void writeMEDITSizeField(Hessian *hess,pMesh mesh );
  void writeMEDITSolution(pMesh mesh);


  void
  phastaTransfer( MeshModType mtype, pMeshChanges mco, void *userData); 


  int 
  inverseMap( pRegion region, 
            double gpt[3],
            double xisol[3] ) ;


  double* 
  InterpolateSolution( pRegion region, 
                     double xi[3], 
                     int ndof,
                     pMeshDataId modes ) ;

   void
   fix4SolutionTransfer(pMesh mesh);



  void 
  R_entitiesAdapt( pRegion region, 
            pVertex *vrts, 
            pEdge *edgs,
            pFace *fcs ) ;

  // for debug
  void check(pMesh);

void diffVt(double *a,double *b,double *v);
double dotProd(double *v1, double *v2);
void crossProd(double *v1, double *v2, double *cp);
int normVt(double *v1,double *nv)  ;

#ifdef __cplusplus
extern "C" {
#endif

// for solving linear system (small)
// the last array is the right hand side
// it is being passed as a reference  and overridden
// to contain the linear system's solution !
#ifndef CYGWIN
  void ludcmp_(double*, int*, int*, int*,double*);
#define ludcmp ludcmp_
   void lubksb_(double*, int*, int*, int*,double*);
#define lubksb lubksb_
   void mytred_(int*,int*,double[3][3],double*,double*,double[3][3]);
#define mytred mytred_  
   void tql2_(int*,int*,double*,double*,double[3][3],int*);
#define tql2 tql2_    
#else
   void LUDCMP(double*, int*, int*, int*,double*);
   void LUBKSB(double*, int*, int*, int*,double*);
   void MYTRED(int*,int*,double[3][3],double*,double*,double[3][3]);
   void TQL2(int*,int*,double*,double*,double[3][3],int*);
#define ludcmp LUDCMP
#define lubksb LUBKSB
#define mytred MYTRED
#define tql2 TQL2
#endif
#ifdef __cplusplus
}
#endif




#endif
