#include <fstream>
#include "phAdapt.h"
#include "attachData.h"

using namespace std;
// using std::cout;
// using std::endl;



extern pMeshDataId phasta_solution;
extern pMeshDataId errorIndicatorID;
extern pMeshDataId nodalgradientID;
extern pMeshDataId nodalhessianID;

void setSizeFieldUsingHessians(pMesh mesh,
			       pMSAdapt simAdapter,
			       double factor,
			       double hmax,
			       double hmin,
			       int option)
{
  int nshg=M_numVertices(mesh);  // only true for linear elements
  double tol=1.e-12;
  
  double T[3][3];
  double maxEigenval=0;
  int i,j;
  
  pVertex vertex;
  double eloc;  	  // local error at a vertex
  double etot=0.;	  // total error for all vertices
  double emean; 	  // emean = etot / nv
  double elocmax=0.;	  // max local error
  double elocmin=1.e20;   // min local error

  std::ofstream adaptSimLog("phAdapt.log");

  // struct Hessian contains decomposed values
  // mesh sizes and directional information
  Hessian *hess = new Hessian[nshg];

  VIter vit=M_vertexIter(mesh);
  i=0;
  while(vertex=VIter_next(vit)) { 
    double xyz[3];
    V_coord(vertex,xyz);

    // get hessian of appropriate variable at a `vertex'
    if(option==1) 
      V_Hessian(vertex,T);
    else
      V_AnalyticHessian(vertex,T,option);
 
    // `T' is the actual Hessian 
    // of any solution variable (e.g., T, u, v, w etc.)
    // `hess' has the information related to Mesh-Metric (eventually)
    // used for mapping the ellipsoid into a unit sphere 
    
    // compute eigen values and eigen vectors
//      int nbEigen = eigen(T,hess[i].dir,hess[i].h);

//      cout<<"eigenvals/eigenVecs (eigen-Li)\n";
//      for(int j=0; j< 3; j++){

//          cout<<hess[i].h[j]<<"\n";
//          cout<<hess[i].dir[0][j]<<" "<<hess[i].dir[1][j]<<" "<<hess[i].dir[2][j]<<"\n";
//      }
        


//      if(nbEigen == 1) {
//        printf(" Vertex Number : %d\n",i);
//        printf(" xyz : %f %f %f \n",xyz[0],xyz[1],xyz[2]);
//        printf("       %f %f %f %f %f %f\n\n",T[0][0], 
//               T[0][1],T[0][2],T[1][1],T[1][2],T[2][2]);
//      }
//      cout<<"Hessian\n";
    
//      for(int j=0; j< 3; j++){
//              cout<<T[j][0]<<" "<<T[j][1]<<" "<<T[j][2]<<"\n";
//      }

    double eigenVals[3];
    double e[3];
    

    // copy T into temporary buffer
    double Tfoo[3][3];
    for(int j=0; j< 3; j++){
        for(int k=0; k< 3; k++){       
            
            Tfoo[j][k]=T[j][k];
        }
    } 

    int three = 3;
    double z[3][3];
    mytred(&three,&three,Tfoo,eigenVals,e,z);
//    cout<<"eigenvals/eigenVecs (tred)\n";

    int ierr;
    tql2(&three,&three,eigenVals,e,z,&ierr);


    for(int j=0; j< 3; j++){
        hess[i].h[j]=eigenVals[j];
        for(int k=0; k< 3; k++){   
            hess[i].dir[j][k]=z[j][k];
        }
    }
//      for(int j=0; j< 3; j++){

//          cout<<hess[i].h[j]<<"\n";
//          cout<<hess[i].dir[0][j]<<" "<<hess[i].dir[1][j]<<" "<<hess[i].dir[2][j]<<"\n";
//      }
        
    
           

    hess[i].h[0] = ABS(hess[i].h[0]);
    hess[i].h[1] = ABS(hess[i].h[1]);
    hess[i].h[2] = ABS(hess[i].h[2]);
    
    if( MAX(hess[i].h[0],MAX(hess[i].h[1],hess[i].h[2])) < tol ) {
      printf("Warning: zero maximum eigenvalue for node %d !!!\n",i);
      printf("       %f %f %f\n", hess[i].h[0],
             hess[i].h[1],hess[i].h[2]);

      adaptSimLog<<"Error: zero maximum eigenvalue !!!"<<endl;
      adaptSimLog<<hess[i].h[0]<<" "<<hess[i].h[1]<<" "<<hess[i].h[2]<<endl;
      adaptSimLog<<T[0][0]<<" "<<T[0][1]<<" "<<T[0][2]<<" ";
      adaptSimLog<<T[1][1]<<" "<<T[1][2]<<" "<<T[2][2]<<endl;
      // exit(1);
      continue;
    }

    // estimate relative interpolation error
    // needed for scaling metric field (mesh size field)
    // to get an idea refer Appendix A in Li's thesis
    eloc=maxLocalError(vertex,T);
    etot += eloc;
    if( eloc>elocmax )  elocmax=eloc;
    if( eloc<elocmin )  elocmin=eloc;

    ++i;
  }
  printf("Info: Reading hessian... done...\n");

  emean =  etot / nshg;
  printf("\n Info on relative interpolation error: ");
  printf("   total: %f\n",etot);
  printf("   mean:  %f\n",emean);
  printf("   max local: %f\n",elocmax);
  printf("   min local: %f\n",elocmin);

  eloc = emean*factor;

  printf("\n towards uniform local error distribution of %f\n", eloc);
  printf("   with max edge length=%f; min edge length=%f\n\n",hmax,hmin);

  adaptSimLog<<"MeshSim   Library build ID : "<<SimMeshing_buildID()<<endl;
  adaptSimLog<<"MeshTools Library build ID : "<<SimMeshTools_buildID()<<endl;
  adaptSimLog<<endl;

  adaptSimLog<<"Strategy chosen is anisotropic adaptation, i.e., size-field driven"<<endl;
  if(option==1)
    adaptSimLog<<"Using numerical/computed hessian... done..."<<endl;
  else
    adaptSimLog<<"Using analytic hessian... done..."<<endl;
  adaptSimLog<<"Info on relative interpolation error :"<<endl;
  adaptSimLog<<"total : "<<etot<<endl;
  adaptSimLog<<"mean : "<<emean<<endl;
  adaptSimLog<<"factor : "<<factor<<endl;
  adaptSimLog<<"min. local : "<<elocmin<<endl;
  adaptSimLog<<"max. local : "<<elocmax<<endl;
  adaptSimLog<<"towards uniform local error distribution of "<<eloc<<endl;
  adaptSimLog<<"with min. edge length : "<<hmin<<endl;
  adaptSimLog<<"with max. edge length : "<<hmax<<endl;

  int foundHmin = 0;
  int foundHmax = 0;
  int hminCount = 0;
  int hmaxCount = 0;
  int bothHminHmaxCount = 0;
  double tol2, tol3;
  // scale the hessian
  // double errorSqRoot = hmin*sqrt(maxEigenval);
  VIter_reset(vit);
  i=0;
  while ( vertex=VIter_next(vit)) {
    tol2 = 0.01*hmax;
    tol3 = 0.01*hmin;
    foundHmin = 0;
    foundHmax = 0;
    for( j=0; j<3; j++ ) {
      if( hess[i].h[j] < tol )
        hess[i].h[j] = hmax;
      else {
        hess[i].h[j] = sqrt(eloc/hess[i].h[j]);
        if( hess[i].h[j] > hmax )
          hess[i].h[j] = hmax;
        if( hess[i].h[j] < hmin )
          hess[i].h[j] = hmin;
      }
    }

    for(j=0; j<3; j++) {
      if(ABS(hess[i].h[j]-hmax) <= tol2)
	foundHmax = 1;
      if(ABS(hess[i].h[j]-hmin) <= tol3)
	foundHmin = 1;
    }
    if(foundHmin)
      hminCount++;
    if(foundHmax)
      hmaxCount++;
    if(foundHmin && foundHmax)
      bothHminHmaxCount++;

    // Modify mesh-metric to take special care
    // like in parallel plates (i.e, w=f(y)) 
    // if a user wants different hmax in x-direction
    // other examples can be situations where
    // user don't want to have one mesh edge 
    // connecting two geometric model edges (like no dofs)
    // ModifyMetric(vertex,hess[i].dir,hess[i].h);

    // set the data in MeshSim format
    for (int jRow=0; jRow<3; jRow++) {
      for(int iDir=0; iDir<3; iDir++) {
        hess[i].dir[jRow][iDir]=hess[i].dir[jRow][iDir]*hess[i].h[jRow];
      }
    }

    MSA_setAnisoVertexSize(simAdapter, 
        		   vertex,
        		   hess[i].dir);
//      cout<<"SIZE node "<<i<<"\n";
//      for(int j=0; j< 3; j++)
//         cout<<hess[i].dir[j][0]<<" "<<hess[i].dir[j][1]<<" "<<hess[i].dir[j][2]<<"\n";
//      cout<<"\n";
       

    ++i;
  }
  VIter_delete(vit);

  cout<<"Nodes with hmin into effect : "<<hminCount<<endl;
  cout<<"Nodes with hmax into effect : "<<hmaxCount<<endl;
  cout<<"Nodes with both hmin/hmax into effect : "<<bothHminHmaxCount<<endl;
  cout<<endl;
  
  adaptSimLog<<"Nodes with hmin into effect : "<<hminCount<<endl;
  adaptSimLog<<"Nodes with hmax into effect : "<<hmaxCount<<endl;
  adaptSimLog<<"Nodes with both hmin/hmax into effect : "<<bothHminHmaxCount<<endl;
  adaptSimLog.close();

  writeMEDITSizeField(hess,mesh);

  if(option==1) {
    cleanAttachedData(mesh,nodalgradientID,0);
    cleanAttachedData(mesh,nodalhessianID,0);
    
    MD_deleteMeshDataId(nodalgradientID);
    MD_deleteMeshDataId(nodalhessianID);
  }

  delete [] hess;
}

// max relative interpolation error at a vertex
double maxLocalError(pVertex vertex, double H[3][3])
{
  pEdge edge;
  double locE;
  double maxLocE=0;

  for(int i=0; i<V_numEdges(vertex); i++ ) {
    edge=V_edge(vertex,i);
    locE=E_error(edge,H);
    if( locE > maxLocE )
      maxLocE=locE;
  }

  return maxLocE;
}

// relative interpolation error along an edge
double E_error(pEdge edge, double H[3][3])
{
  double xyz[2][3], vec[3];
  double locE=0;
  int i,j;

  V_coord(E_vertex(edge,0),xyz[0]);
  V_coord(E_vertex(edge,1),xyz[1]);
  diffVt(xyz[0],xyz[1],vec);

  for( i=0; i<3; i++ )
    for( j=0; j<3; j++ ) 
      locE += H[i][j]*vec[i]*vec[j];

  return ABS(locE);
}


void setIsotropicSizeField(pMesh mesh,
			   pMSAdapt simAdapter,
			   double factor,
			   double hmax, 
			   double hmin,
			   int option)
{
  // for 3D problems
  int dim = 3;
  // assuming linear basis
  int poly_order = 1;

  double totalError = 0.;
  double threshold  = 0.;
  double sumOfError = 0.;

  pVertex vertex;
  VIter vIter = M_vertexIter(mesh);
  while(vertex = VIter_next(vIter)) {
    double *nodalValue;
    if(EN_getDataPtr((pEntity)vertex,errorIndicatorID,(void**)&nodalValue) == NULL) {
      cout<<"\nerror in setIsotropicSizeField(...) : no error data attached to vertex"<<endl;
      exit(0);
    }

    // eta_domain^2 = sum_k (eta_k^2)
    totalError += *nodalValue;
    // eta_sum = sum_k (eta_k)
    sumOfError += sqrt(*nodalValue);
  }

  totalError = sqrt(totalError);

  std::ofstream sizes("isoSize.dat");  
  std::ofstream adaptFactorFile("adaptFactor.dat");

  VIter_reset(vIter);  
  while(vertex = VIter_next(vIter)) {
    double *nodalValue;
    if(EN_getDataPtr((pEntity)vertex,errorIndicatorID,(void**)&nodalValue) == NULL) {
      cout<<"\nerror in setIsotropicSizeField(...) : no error data attached to vertex"<<endl;
      exit(0);
    }

    double oldSize = 0.;
    pEdge edge;
    int numEdges = V_numEdges(vertex);
    // old size at a location (vertex)
    // is mean value of the lengths 
    // of edges around (seems ok for isotropic meshes)
    // (can have different choices)
    for (int i=0; i < numEdges; i++) {
      edge = V_edge(vertex,i);
      oldSize += E_length(edge);
    }
    oldSize = oldSize/numEdges;

    threshold = factor*totalError;
    // based on mesh optimal criteria and error convergence rate
    // implemented for linear cases
    // refer L.-Y. Li, Comm. Num. Meth. Eng.,
    // Vol. 11, 857-868 & 911-915 (1995)
    double adaptFactor = threshold/(pow(*nodalValue,0.25)*sqrt(sumOfError));

    double MaxRefineFactor = hmin;
    double MaxCoarsenFactor = hmax;

    // these factors set cut-off levels for adaptation
    // can set such that no coarsening (or refinement) occurs
    if (MaxRefineFactor > 1.e-10  && adaptFactor < 1.0/MaxRefineFactor) 
      adaptFactor = 1.0/MaxRefineFactor;

    if (MaxCoarsenFactor > 1.e-10 && adaptFactor > MaxCoarsenFactor) 
      adaptFactor = MaxCoarsenFactor;

    double newSize = adaptFactor*oldSize;

    MSA_setVertexSize(simAdapter, 
		      vertex,
		      newSize);

    sizes<<newSize<<"\n";
    adaptFactorFile<<adaptFactor<<"\n";
  }
  VIter_delete(vIter);

  sizes.close();
  adaptFactorFile.close();

  cout<<"Info. on adaptation parameters are: "<<endl;
  cout<<"Total Error      : "<<totalError<<endl;
  cout<<"factor           : "<<factor<<endl;
  cout<<"Threshold        : "<<threshold<<endl;
  cout<<"sumOfError       : "<<sumOfError<<endl;
  cout<<"MaxCoarsenFactor : "<<hmax<<endl;
  cout<<"MaxRefineFactor  : "<<hmin<<"\n"<<endl;

  std::ofstream adaptSimLog("phAdapt.log");

  adaptSimLog<<"Strategy chosen is size-field driven for isotropic adaptation"<<endl;  
  adaptSimLog<<"Info. on adaptation parameters are: "<<endl;
  adaptSimLog<<"Total Error      : "<<totalError<<endl;
  adaptSimLog<<"factor           : "<<factor<<endl;  
  adaptSimLog<<"Threshold        : "<<threshold<<endl;
  adaptSimLog<<"sumOfError       : "<<sumOfError<<endl;
  adaptSimLog<<"MaxCoarsenFactor : "<<hmax<<endl;
  adaptSimLog<<"MaxRefineFactor  : "<<hmin<<endl;
  adaptSimLog.close();
}


void 
setManualSizeField(pMesh mesh,
		   pMSAdapt simAdapter, 
        	   int strategy) {
  double dir[3][3];
  char option[28];
  
  switch(strategy) {
  case -1:
    {
      sprintf(option,"constant");

      dir[0][0]=0.1;//x-direction
      dir[1][1]=0.1;//y-direction
      dir[2][2]=6.0;//z-direction
    
      dir[0][1]=dir[0][2]=0.;
      dir[1][0]=dir[1][2]=0.;
      dir[2][0]=dir[2][1]=0.;

      pVertex vertex;
      VIter vit=M_vertexIter(mesh);
      while(vertex=VIter_next(vit)) {   
	MSA_setAnisoVertexSize(simAdapter, 
			       vertex,
			       dir);
      }
      VIter_delete(vit);
    }
    break;
  case -2:
    {
      sprintf(option,"cylindrical");

      double norm,tol=1.e-8;
    
      double sizeR=0.1;
      double sizeTheta=0.1;
      double sizeZ=1.0;

      pVertex vertex;
      VIter vit=M_vertexIter(mesh);
      while(vertex=VIter_next(vit)) {  
	double xyz[3];
	V_coord(vertex,xyz);
	norm=sqrt(xyz[0]*xyz[0]+xyz[1]*xyz[1]);
	if( norm>tol ) 
	  {
	    dir[0][0]=sizeR*(xyz[0]/norm);
	    dir[0][1]=sizeR*(xyz[1]/norm);
	    dir[0][2]=0.;
	    dir[1][0]=-sizeTheta*(xyz[1]/norm);
	    dir[1][1]=sizeTheta*(xyz[0]/norm);
	    dir[1][2]=0.;
	    dir[2][0]=0.;
	    dir[2][1]=0.;
	    dir[2][2]=sizeZ;
	  }
	else
	  {
	    dir[0][0]=sizeR;
	    dir[0][1]=0.;
	    dir[0][2]=0.;
	    dir[1][0]=0.;
	    dir[1][1]=sizeTheta;
	    dir[1][2]=0.;
	    dir[2][0]=0.;
	    dir[2][1]=0.;
	    dir[2][2]=sizeZ;
	  }

	MSA_setAnisoVertexSize(simAdapter, 
			       vertex,
			       dir);
      }
      VIter_delete(vit);
    }
    break;
  default :
    cout<<"check strategy [in setManualSizefield(...)]"<<endl;
    exit(-1);
    break; 
  }

  std::ofstream adaptSimLog("phAdapt.log");
  adaptSimLog<<"Strategy chosen for adaptation is size-field driven"<<endl;
  adaptSimLog<<"Mesh size-field is set manually"<<endl;
  adaptSimLog<<"Size-field option : "<<option<<endl;
  adaptSimLog.close();
}

// utilities

void diffVt(double *a,double *b,double *v)
{
  v[0] = a[0] - b[0] ;
  v[1] = a[1] - b[1] ;
  v[2] = a[2] - b[2] ;
}



/*----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
double dotProd(double *v1, double *v2)
{
  return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2] ;
}

/*----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void crossProd(double *v1, double *v2, double *cp)
{
  cp[0] = v1[1]*v2[2] - v1[2]*v2[1] ;
  cp[1] = v1[2]*v2[0] - v1[0]*v2[2] ;
  cp[2] = v1[0]*v2[1] - v1[1]*v2[0] ;
}

/*----------------------------------------------------------------------------
 normalize a 3D vector
----------------------------------------------------------------------------*/
int normVt(double *v1,double *nv)
{
  double norm ;

  norm = v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2] ;


  norm = 1./sqrt(norm) ;
  nv[0] = v1[0]*norm ; 
  nv[1] = v1[1]*norm ; 
  nv[2] = v1[2]*norm ;

  return(1) ;
}


