#include <iostream>
#include <string>
#include "phAdapt.h"
#include "phastaIO.h"
#include <time.h>
#include <stdio.h>

#ifndef CYGWIN
#include <unistd.h>
#include <strings.h>
#endif



using namespace std;



extern pMeshDataId nodalhessianID;

/////////////////////////////////
// write out the restart files //
// which contain the hessians  //
/////////////////////////////////
void
writeRestartHessians(pMesh mesh )
{
  double* nodalHessian;
  pVertex vertex;
  VIter vIter = M_vertexIter( mesh );
  // variables written out
  int nshg_fine = M_numVertices(mesh);
  double* q = new double [ nshg_fine * 6 ];    
  int vcounter=0;
  
  while ( vertex = VIter_next( vIter ) ) {
    if(EN_getDataPtr( (pEntity)vertex, nodalhessianID,
		      (void**)&nodalHessian)  != NULL) {
      int j=0;
      for( int i=0; i<6; i++ ) {	
	// store the hessians into q vector
	q[vcounter+j] = nodalHessian[i]; 
	
	j=j+nshg_fine;
      }
      vcounter = vcounter+1;  
      
    }
    // there is no hessian on this vertex
    else{
      cerr<<"\nerror in writeRestartHessians: encountered\n"
	  <<"a vertex that carries no hessian\n";
      exit(0);
    }    
  }
  VIter_delete(vIter); 
  
  // write the files
  char fname[255];
  char* oformat ="binary";

  int magic_number = 362436;
  int* mptr = &magic_number;
  int  frest;
  int iarray[10];
  int size, nitems;
  char filename[255];
  
  openfile_( "restart.1111.1", "write", &frest );
  
  writestring_( &frest,"# PHASTA Input File Version 2.0\n");
  writestring_( &frest, "# Byte Order Magic Number : 362436 \n");
  sprintf(fname,"# Output generated by phAdapt version: %s \n");
  
  writestring_( &frest, fname );
 
  time_t timenow = time ( &timenow);

  sprintf(fname,"# %s\n", ctime( &timenow ));
  
  writestring_( &frest, fname );
  
  int one=1;
  
  size = 1;
  nitems = 1;// length of iarray
  iarray[ 0 ] = 1;
  
  writeheader_( &frest, "byteorder magic number ",
                  (void*)iarray, &nitems, &size, "integer", oformat );

  writedatablock_( &frest, "byteorder magic number ",
                     (void*)mptr, &nitems, "integer", oformat );
  
  /////////////////////////
  // writing the restart //
  /////////////////////////

  bzero( (void*)fname, 255 );
  sprintf(fname,"number of modes : < 0 > %d\n", nshg_fine);
  writestring_( &frest, fname );
  
  bzero( (void*)fname, 255 );
  sprintf(fname,"number of variables : < 0 > %d\n", 6);
  writestring_( &frest, fname );
  
  size =  6*nshg_fine;
  nitems = 3; // length of iarray
  iarray[0] = nshg_fine;
  iarray[1] = 6;
  iarray[2] = 1111; // stepNumber;
  
  writeheader_( &frest, "solution ",
		( void* )iarray, &nitems, &size,"double", oformat );
  
  nitems = 6*nshg_fine;
  writedatablock_( &frest, "solution ",
		   ( void* )(q), &nitems, "double", oformat );
  
  closefile_( &frest, "write" );
  // finished writing the restart //
}

