#ifdef FMDB

#include "phParAdapt.h"
#include <iostream>
#include "MA.h"

extern double nuYbyRho;
//the variable nuYbyRho is actually nu*Y1plus/sqrt(rho)
//this will be an input parameter passed from adapt.inp
//Y1plus value for RANS(spallart Almaras = 1.0-5.0, 1.0 is better
//Y1plus value for k-eps and k-omega models 0.1-0.5
//Y1plus for Wall model 30-50  (refer literature for best suitable value)

extern pMeshDataId wallStressID;

void DistOffWall(pMesh mesh, pMSAdapt simAdapter){

   pVertex vertex;
   VIter vIter=M_vertexIter(mesh);
   int vCount=0;

   while(vertex = VIter_next(vIter)){
      int Gtype; 
      double *Tau; 
      double uTau;
      double WallDist;
      Gtype = V_whatInType(vertex);
// check if the vertex lies on model boundary and if owned by the proc      
      if(Gtype==0 || Gtype==1 || Gtype==2 
            && EN_isOwnerProc((pEntity)vertex)){
         if(!EN_getDataPtr(vertex, wallStressID, (void**)&Tau)){
            cout<<"\nError in DistOffWall, no wall shear stress data attached to vertex\n";
            exit(0);
            V_info(vertex);
         }
      uTau=0.0;   
      for(int i=0; i<6; i++){
         uTau = uTau + Tau[i]*Tau[i];
      }  //assembling the dot product of the wall shear stress vector

      uTau = sqrt(ABS(sqrt(uTau)));
      
      WallDist = nuYbyRho/uTau; //nu*Y1plus/(sqrt(rho)*uTau)
//insert a limiting condition when uTau tends to be zero, since that would
//make WallDist fairly large and we can not accept that.
//calculation reqd to compare this distance to other 2 in plane directions
//WallDist value should ALWAYS be less than any other delta value
//to have acceptable aspect ratio.
      
      MA_SetFirstLayerThickness(simAdapter, vertex, WallDist);
      }
   }
   VIter_delete(vIter);     

}

#endif
