#include <iostream>
#include "phastaIO.h"
#include "MeshSim.h"
#include "shapeFunction.h"
using namespace std;

extern int

restart( int irestart, double* q, int nshg, int nvr, char fieldname[]);
void solution( pEntity ent, pMeshDataId phSol, double** sol );
void numberofmodes( pEntity ent, pMeshDataId modes, int* sol );
int inverseMap( pRegion region, double* qpt, double* pt ) ;
void R_entities( pRegion region, pVertex *vrts, pEdge *edgs,pFace *fcs ) ;

int
attachSolution( char *solfile, 
                pMesh mesh, 
                pMeshDataId phSol,
                pMeshDataId modes,
                char fieldname[],
                int ndof ) {
    
    double* soln;
    int ione =1;
    int iarray[2]={-1,-1};

    int irestart;
    openfile_( solfile , "read",  &irestart );

    readheader_( &irestart, "number of modes", iarray, &ione, "integer", "binary");
    int nshg = iarray[0];
    double* q_ji = new double[nshg*ndof];
    // read solution file
    int stepno = restart( irestart, q_ji, nshg, ndof, fieldname);
    double* q = new double[nshg*ndof];
    
    // transpose the solution
    // q_ji = [ndof][nshg]
    // q    = [nshg][ndof]
    
    for(int i = 0; i< nshg; i++)
        for( int j=0; j< ndof; j++)
            q[i*ndof+j] = q_ji[j*nshg+i];
    
    delete [] q_ji;
    
    
    /* attach the vertex coefficients */
    pVertex vertex;
    VIter vIter = M_vertexIter( mesh );
    while ((vertex = VIter_next( vIter ) )) {
        soln = new double[ndof];
        for (int i=0; i < ndof; i++) 
            soln[i] = q[ndof*(EN_id( (pEntity)vertex)) + i ];
        EN_attachDataPtr( (pEntity)vertex, phSol, (void *) soln );
    }
    VIter_delete( vIter );


    //  now we need to read the mapping between sms and phasta dofs for
    //  edges and faces. 

    iarray[0]=-1;
    iarray[1]=-1;
    int itwo=2;
    readheader_( &irestart, "edge mode mapping to sms", iarray, &itwo,
                 "integer", "binary");
    int size = iarray[0];
    int nmodes = iarray[0];
    if ( size > 0 ) {
        int* rmap = new int [ size ];
        readdatablock_( &irestart, "edge mode mapping to sms", rmap,
                        &size, "integer", "binary");
        int* ptr = rmap;
        int nem;
        pEdge edge;
        int ecounter =0;
        EIter eIter = M_edgeIter( mesh );
        while ( (edge = EIter_next( eIter )) && (ecounter < nmodes )) {
            if ( EN_id( (pEntity)edge ) == *(ptr) ) {
                ptr++;
                nem = *(ptr++);
                if ( nem > 0 ) {
                    ecounter += nem;
                    soln = new double[ndof*nem];
                    for(int i=0; i < nem ; i++ ) {
                        for (int j=0; j < ndof; j++) 
                            soln[i*ndof+j] = q[(*ptr)*ndof + j];
                        ptr++;
                    }
                    EN_attachDataPtr( (pEntity)edge, phSol, (void *) soln );
                    EN_attachDataInt( (pEntity)edge, modes, nem );
                }
            }
        }
        EIter_delete( eIter );
        delete [] rmap;
    }
    
    iarray[0]=-1;
    iarray[1]=-1;
    readheader_( &irestart, "face mode mapping to sms", iarray, &itwo,
                 "integer", "binary");
    size = iarray[0];
    nmodes = iarray[0];

    if ( size > 0 ) {
        int* rmap = new int [ size ];
        readdatablock_( &irestart, "face mode mapping to sms", rmap,
                        &size, "integer", "binary");
        int* ptr = rmap;
        int nem;
        pFace face ;
        int fcounter=0;
        FIter fIter = M_faceIter( mesh );
        while( ( face = FIter_next( fIter ) ) && ( fcounter < nmodes )) {
            if ( EN_id( (pEntity)face ) == *(ptr) ) {
                ptr++;
                nem = *(ptr++);
                if ( nem > 0 ) {
                    if ( nem > 1 )  {
                        cerr << "face cannot have 2 modes at the moment" << endl;
                        cerr << "please check  :" << __FILE__ << endl;
                        exit(0);
                    }
                    fcounter += nem;
                    soln = new double[ndof*nem];
                    for(int i=0; i < nem ; i++ ) {
                        for (int j=0; j < ndof; j++) 
                            soln[i*ndof+j] = q[(*ptr)*ndof + j];
                        ptr++;
                    }
                    EN_attachDataPtr( (pEntity)face, phSol, (void *) soln );
                    EN_attachDataInt( (pEntity)face, modes, nem );
                }
            }
        }
        FIter_delete( fIter );
        delete[] rmap;
    }
        
    closefile_( &irestart, "read");

    delete[] q;

    return stepno;
}

double*
getSolution( double* pt,
             pRegion region,
             pMeshDataId phSol,
             pMeshDataId modes,
             double* params = NULL,
             int ndof = 5) 
{

    double xi[3] ;
    double* q = NULL;

    // can't trust the xietazeta provided by simmetrix 
    // talk to onkar abt this.
    /*
    if ( params != NULL ) {
        xi[0] = params[0];
        xi[1] = params[1];
        xi[2] = params[2];
    }
    */

    if ( inverseMap( region, pt , xi ) ) {
        double* vcc;
        double* ecc;
        double* fcc;
        double r=xi[0];
        double s=xi[1];
        double t=xi[2];
        double L[4] = {r,s,t,1-r-s-t};

        q = new double[ndof+1];
        for (int i=0; i < ndof; i++) q[i] = 0.0;
        q[ndof]=1.0;

        pVertex vrts[4];
        pEdge edgs[6];
        pFace facs[4];

        R_entities(region,vrts,edgs,facs);

        for (int k=0; k < 4; k++) {
            solution(( pEntity )vrts[k], phSol, &vcc );
            double sfval = Entity_shapeFunction(( pEntity )vrts[k], ( pEntity )region, 1, 0, L);
            for (int i=0; i < ndof; i++) {
                q[i] += vcc[i]*sfval;
            }
        }

        // get the edge and coefficients
        for (int k=0; k < 6; k++) {   
            int nem=0;   
            solution(( pEntity )edgs[k], phSol, &ecc );   
            numberofmodes((pEntity)edgs[k], modes, &nem );   
            if ( ecc ) {   
                for (int e=0; e < nem ; e++) {   
                    double sfval = Entity_shapeFunction(( pEntity )edgs[k],( pEntity )region,e+2,0,L);   
                    for (int j=0; j < ndof; j++) {   
                        q[j] += ecc[e*ndof+j]*sfval;   
                    }   
                }   
            }   
            double fnem = nem*1.0;
            if ( fnem > ( q[ndof]-1 ) ) q[ndof] += 1.0;
        }   
        for (int k=0; k < 4; k++) {   
            solution(( pEntity )facs[k], phSol, &fcc );   
            if ( fcc ) {
                /* currently only p=3 */
                double sfval =  Entity_shapeFunction(( pEntity )facs[k],( pEntity )region,3,0,L);   
                for (int j=0; j < ndof; j++) {   
                    q[j] += fcc[j]*sfval;   
                }   
                q[ndof] = 3.0;
            }
        }   
        return q;
    }
}
