#include <stdio.h>
#include <iostream>
#include "MeshSim.h"
#include "MeshSimInternal.h"
#include <map>
using namespace std;

extern "C"
void restart( char* solfile,
              double* q,
              int nshg,
              int nvr );

void attachSolution(char* solfile, 
                    pMesh mesh, 
                    map<pEntity, double *>& data,
		            int ndof, 
                    int P ) {
  int count,i;
  double *soln;
  int nem = (P > 1) ? (P - 1) : 0;
  int nfm = (P > 3) ? ((P-3)*(P-2)/2.0) : 0;
  int nrm = (P > 5) ? ((P-4)*(P-5)*(P-3)/6.0) : 0;
  
  int nshg = M_numVertices(mesh) + nem * M_numEdges(mesh)
    + nfm * M_numFaces(mesh) + nrm * M_numRegions(mesh);

  double* q_ji = new double[nshg*ndof];
  double* q = new double[nshg*ndof];
  
  // read solution file

  restart( solfile, q_ji, nshg, ndof );

  // transpose the solution

  for(i = 0; i< nshg; i++)
      for( int j=0; j< ndof; j++)
          q[i*ndof+j] = q_ji[j*nshg+i];

  delete [] q_ji; 
  
  /* attach the vertex coefficients */
  count = 0;
  pVertex vertex;
  VIter vIter = M_vertexIter( mesh );
  while (vertex = VIter_next( vIter ) ) {
    soln = new double[ndof];
    for (i=0; i < ndof; i++) soln[i] = q[count++];
    data[(pEntity)vertex]=soln;
  }
  VIter_delete( vIter );
  
  /* attach the edge coefficients */
  if (nem > 0){
    pEdge edge;
    EIter eIter = M_edgeIter( mesh );
    while (edge = EIter_next( eIter ) ) {
      soln = new double[ndof*nem];
      for (i=0; i < ndof*nem; i++) soln[i] = q[count++];
      data[(pEntity)edge]=soln;
    }
    EIter_delete( eIter );
  }
  
  /* attach face coefficients */
  if (nfm > 0){
    pFace face;
    FIter fIter = M_faceIter( mesh );
    while (face = FIter_next( fIter ) ) {
      soln = new double[ndof*nfm];
      for (i=0; i < ndof*nfm; i++) soln[i] = q[count++];
      data[(pEntity)face]=soln;
    }
    FIter_delete( fIter );
  }
  /* attach region coefficients */
  if (nrm > 0){
     cerr << " No code to attach Region Modes " << endl;
     exit ( -1 );
  }

  delete [] q;
}
void attachSoln( char *solfile, 
                 pMesh mesh, 
                 pMeshDataId phSol,
                 int ndof, 
                 int P ) {
  int count,i;
  double *soln;
  int nem = (P > 1) ? (P - 1) : 0;
  int nfm = (P > 3) ? ((P-3)*(P-2)/2.0) : 0;
  int nrm = (P > 5) ? ((P-4)*(P-5)*(P-3)/6.0) : 0;
  
  int nshg = M_numVertices(mesh) + nem * M_numEdges(mesh)
    + nfm * M_numFaces(mesh) + nrm * M_numRegions(mesh);

  double* q_ji = new double[nshg*ndof];
  double* q = new double[nshg*ndof];
  
  // read solution file

  restart( solfile, q_ji, nshg, ndof );

  // transpose the solution

  for(i = 0; i< nshg; i++)
      for( int j=0; j< ndof; j++)
          q[i*ndof+j] = q_ji[j*nshg+i];

  delete [] q_ji;

  /* attach the vertex coefficients */
  count = 0;
  pVertex vertex;
  VIter vIter = M_vertexIter( mesh );
  while (vertex = VIter_next( vIter ) ) {
    soln = new double[ndof];
    for (i=0; i < ndof; i++) soln[i] = q[count++];
    EN_attachDataPtr( (pEntity)vertex, phSol, (void *) soln );
  }
  VIter_delete( vIter );

  /* attach the edge coefficients */
  if (nem > 0){
    pEdge edge;
    EIter eIter = M_edgeIter( mesh );
    while (edge = EIter_next( eIter ) ) {
      soln = new double[ndof*nem];
      for (i=0; i < ndof*nem; i++) soln[i] = q[count++];
      EN_attachDataPtr( (pEntity)edge, phSol, (void *) soln );
    }
    EIter_delete( eIter );
  }
  
  /* attach face coefficients */
  if (nfm > 0){
    pFace face;
    FIter fIter = M_faceIter( mesh );
    while (face = FIter_next( fIter ) ) {
      soln = new double[ndof*nfm];
      for ( i=0; i < ndof*nfm; i++ ) soln[i] = q[count++];
      EN_attachDataPtr( (pEntity)face, phSol, (void *) soln );
    }
    FIter_delete( fIter );
  }
  /* attach region coefficients */
  if (nrm > 0) {
     cerr << " No code to attach Region Modes " << endl;
     exit ( -1 );
  }

  delete[] q;
}

void
solution( pEntity ent,
          pMeshDataId phSol,
	  double** sol ) {
  EN_getDataPtr( ent, phSol, (void **)sol );
}

