#include <stdio.h>
#include <iostream>
#include <vector>
#include <map>
#include "phastaIO.h"
#include "MeshSim.h"
using namespace std;

extern int
restart( char* solfile,
         double* q,
         int nshg,
         int nvr );

void attachSolution(char* solfile, 
                    pMesh mesh, 
                    map<pEntity, double *>& data,
		            int ndof, 
                    int P ) {
    int count,i;
    double *soln;
    int nem = (P > 1) ? (P - 1) : 0;
    int nfm = (P > 3) ? ((P-3)*(P-2)/2.0) : 0;
    int nrm = (P > 5) ? ((P-4)*(P-5)*(P-3)/6.0) : 0;
  
    int nshg = M_numVertices(mesh) + nem * M_numEdges(mesh)
        + nfm * M_numFaces(mesh) + nrm * M_numRegions(mesh);

    double* q_ji = new double[nshg*ndof];
    double* q = new double[nshg*ndof];
  
    // read solution file

    restart( solfile, q_ji, nshg, ndof );

    // transpose the solution

    for(i = 0; i< nshg; i++)
        for( int j=0; j< ndof; j++)
            q[i*ndof+j] = q_ji[j*nshg+i];

    delete [] q_ji; 
  
    /* attach the vertex coefficients */
    count = 0;
    pVertex vertex;
    VIter vIter = M_vertexIter( mesh );
    while (vertex = VIter_next( vIter ) ) {
        soln = new double[ndof];
        for (i=0; i < ndof; i++) soln[i] = q[count++];
        data[(pEntity)vertex]=soln;
    }
    VIter_delete( vIter );
  
    /* attach the edge coefficients */
    if (nem > 0){
        pEdge edge;
        EIter eIter = M_edgeIter( mesh );
        while (edge = EIter_next( eIter ) ) {
            soln = new double[ndof*nem];
            for (i=0; i < ndof*nem; i++) soln[i] = q[count++];
            data[(pEntity)edge]=soln;
        }
        EIter_delete( eIter );
    }
  
    /* attach face coefficients */
    if (nfm > 0){
        pFace face;
        FIter fIter = M_faceIter( mesh );
        while (face = FIter_next( fIter ) ) {
            soln = new double[ndof*nfm];
            for (i=0; i < ndof*nfm; i++) soln[i] = q[count++];
            data[(pEntity)face]=soln;
        }
        FIter_delete( fIter );
    }
    /* attach region coefficients */
    if (nrm > 0){
        cerr << " No code to attach Region Modes " << endl;
        exit ( -1 );
    }

    delete [] q;
}
void attachSoln( char *solfile, 
                 pMesh mesh, 
                 pMeshDataId phSol,
                 int ndof, 
                 int P ) {
    int count,i;
    double *soln;
    int nem = (P > 1) ? (P - 1) : 0;
    int nfm = (P > 3) ? ((P-3)*(P-2)/2.0) : 0;
    int nrm = (P > 5) ? ((P-4)*(P-5)*(P-3)/6.0) : 0;
  
    int nshg = M_numVertices(mesh) + nem * M_numEdges(mesh)
        + nfm * M_numFaces(mesh) + nrm * M_numRegions(mesh);

    double* q_ji = new double[nshg*ndof];
    double* q = new double[nshg*ndof];
  
    // read solution file

    restart( solfile, q_ji, nshg, ndof );

    // transpose the solution

    for(i = 0; i< nshg; i++)
        for( int j=0; j< ndof; j++)
            q[i*ndof+j] = q_ji[j*nshg+i];

    delete [] q_ji;

    /* attach the vertex coefficients */
    count = 0;
    pVertex vertex;
    VIter vIter = M_vertexIter( mesh );
    while (vertex = VIter_next( vIter ) ) {
        soln = new double[ndof];
        for (i=0; i < ndof; i++) soln[i] = q[count++];
        EN_attachDataPtr( (pEntity)vertex, phSol, (void *) soln );
    }
    VIter_delete( vIter );

    /* attach the edge coefficients */
    if (nem > 0){
        pEdge edge;
        EIter eIter = M_edgeIter( mesh );
        while (edge = EIter_next( eIter ) ) {
            soln = new double[ndof*nem];
            for (i=0; i < ndof*nem; i++) soln[i] = q[count++];
            EN_attachDataPtr( (pEntity)edge, phSol, (void *) soln );
        }
        EIter_delete( eIter );
    }
  
    /* attach face coefficients */
    if (nfm > 0){
        pFace face;
        FIter fIter = M_faceIter( mesh );
        while (face = FIter_next( fIter ) ) {
            soln = new double[ndof*nfm];
            for ( i=0; i < ndof*nfm; i++ ) soln[i] = q[count++];
            EN_attachDataPtr( (pEntity)face, phSol, (void *) soln );
        }
        FIter_delete( fIter );
    }
    /* attach region coefficients */
    if (nrm > 0) {
        cerr << " No code to attach Region Modes " << endl;
        exit ( -1 );
    }

    delete[] q;
}

int
attachVPSoln( char *solfile, 
              pMesh mesh, 
              pMeshDataId phSol,
              pMeshDataId mark,
              pMeshDataId modes,
              int ndof ) {

    double* soln;
    int refmap =0;
    openfile_( "geombc.dat.1", "read", &refmap );
    if ( refmap == 0 ) {
        cout << "file not there " << endl;
        exit(0);
    }
    int itwo = 2;
    int iarray[3]={0,0,0};
    readheader_( &refmap, "higher order refinement mapping", iarray, &itwo,
                 "integer", "binary");

    int size = iarray[0];
    int nshg = iarray[1];
    int emodes= iarray[2];
    int fmodes= iarray[3];
    int* rmap = new int [ size ];
    double* q_ji = new double[nshg*ndof];
    double* q = new double[nshg*ndof];

    readdatablock_( &refmap, "higher order refinement mapping", rmap,
                    &size, "integer", "binary");
  
    closefile_( &refmap, "read");

    // read solution file

    int stepno = restart( solfile, q_ji, nshg, ndof );

    // transpose the solution
    // q_ji = [ndof][nshg]
    // q    = [nshg][ndof]

    for(int i = 0; i< nshg; i++)
        for( int j=0; j< ndof; j++)
            q[i*ndof+j] = q_ji[j*nshg+i];

    delete [] q_ji;

    /* attach the vertex coefficients */
    pVertex vertex;
    VIter vIter = M_vertexIter( mesh );
    while (vertex = VIter_next( vIter ) ) {
        soln = new double[ndof];
        for (int i=0; i < ndof; i++) 
            soln[i] = q[ndof*(EN_id( (pEntity)vertex)) + i ];
        EN_attachDataPtr( (pEntity)vertex, phSol, (void *) soln );
    }
    VIter_delete( vIter );

    int* ptr = rmap;
    int nem;
    pEdge edge;
    int ecounter =0;
    EIter eIter = M_edgeIter( mesh );
    while ( (edge = EIter_next( eIter )) && (ecounter < emodes )) {
        if ( EN_id( (pEntity)edge ) == *(ptr) ) {
            ptr++;
            nem = *(ptr++);
            if ( nem > 0 ) {
                ecounter += nem;
                soln = new double[ndof*nem];
                for(int i=0; i < nem ; i++ ) {
                    for (int j=0; j < ndof; j++) 
                        soln[i*ndof+j] = q[(*ptr)*ndof + j];
                    ptr++;
                }
                EN_attachDataPtr( (pEntity)edge, phSol, (void *) soln );
                EN_attachDataInt( (pEntity)edge, mark, 1 );
                EN_attachDataInt( (pEntity)edge, modes, nem );
            }
        }
    }
    EIter_delete( eIter );

    pFace face ;
    int fcounter=0;
    FIter fIter = M_faceIter( mesh );
    while( ( face = FIter_next( fIter ) ) && ( fcounter < fmodes )) {
        if ( EN_id( (pEntity)face ) == *(ptr) ) {
            ptr++;
            nem = *(ptr++);
            if ( nem > 0 ) {
                if ( nem > 1 )  {
                    cerr << "face cannot have 2 modes at the moment" << endl;
                    cerr << "please check  :" << __FILE__ << endl;
                    exit(0);
                }
                ecounter += nem;
                soln = new double[ndof*nem];
                for(int i=0; i < nem ; i++ ) {
                    for (int j=0; j < ndof; j++) 
                        soln[i*ndof+j] = q[(*ptr)*ndof + j];
                    ptr++;
                }
                EN_attachDataPtr( (pEntity)face, phSol, (void *) soln );
                EN_attachDataInt( (pEntity)face, mark, 1 );
                EN_attachDataInt( (pEntity)face, modes, nem );
            }
        }
    }
    
    FIter_delete( fIter );
    delete[] q;
    delete[] rmap;

    return stepno;
}

void
solution( pEntity ent,
          pMeshDataId phSol,
	      double** sol ) {
    *sol=0;
    EN_getDataPtr( ent, phSol, (void **)sol );
}

void
numberofmodes( pEntity ent,
               pMeshDataId modes,
	           int* sol ) {
    *sol = 0;
    EN_getDataInt( ent, modes, sol );
}
